@@ -3,7 +3,6 @@ open OUnit
 let all =
   TestList
     [
-      Bin_prot_test.ML.test;
-      Bin_prot_test.C.test;
+      Bin_prot_test.test;
       Bin_prot_test.Common.test;
     ]
